-------------------------------------------------------------------------------
-- Title      : controlUnit
-- Project    :
-------------------------------------------------------------------------------
-- File       : operativeUnit.vhd
-- Author     : Jean-Noel BAZIN  <jnbazin@pc-disi-026.enst-bretagne.fr>
-- Company    :
-- Created    : 2018-04-11
-- Last update: 2019-02-13
-- Platform   :
-- Standard   : VHDL'93/02
-------------------------------------------------------------------------------
-- Description: Control unit of a sequential FIR filter.
-------------------------------------------------------------------------------
-- Copyright (c) 2018
-------------------------------------------------------------------------------
-- Revisions  :
-- Date        Version  Author  Description
-- 2018-04-11  1.0      jnbazin Created
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity controlUnit is

  port (
    I_clock               : in  std_logic;  -- global clock
    I_reset               : in  std_logic;  -- asynchronous global reset
    I_inputSampleValid    : in  std_logic;  -- Control signal to load the input sample in the sample shift register and shift the register
    I_processingDone1     : in  std_logic;
    I_processingDone2_1   : in  std_logic;
    I_processingDone2_2   : in  std_logic;
    I_processingDone3     : in  std_logic;
    O_loadShift           : out std_logic_vector(1 downto 0);  -- filtered sample
    O_initAddress         : out std_logic;  -- Control signal to initialize register read address
    O_incrAddress         : out std_logic;  -- Control signal to increment register read address
    O_initSum             : out std_logic;  -- Control signal to initialize the MAC register
    O_loadSum             : out std_logic;  -- Control signal to load the MAC register;
    O_loadOutput               : out std_logic;  -- Control signal to load Y register
    O_Selector            : out std_logic_vector(1 downto 0)
    --O_FilteredSampleValid : out std_logic  -- Data valid signal for filtered sample
    );

end entity controlUnit;

architecture archi_operativeUnit of controlUnit is


  type T_state is (WAIT_SAMPLE, STORE1, STORE2, STORE3, PROCESSING_LOOP1,PROCESSING_LOOP2_1,PROCESSING_LOOP2_2,PROCESSING_LOOP3, OUTPUT, WAIT_END_SAMPLE);  -- state list
  signal SR_presentState : T_state;
  signal SR_futurState   : T_state;

begin

  process (I_clock, I_reset) is
  begin
    if I_reset = '1' then               -- asynchronous reset (active high)
      SR_presentState <= WAIT_SAMPLE;
    elsif rising_edge(I_clock) then     -- rising clock edge
      SR_presentState <= SR_futurState;
    end if;
  end process;

  process (I_inputSampleValid, I_processingDone1, I_processingDone2_1, I_processingDone2_2, I_processingDone3, SR_presentState) is
  begin
    case SR_presentState is

      when WAIT_SAMPLE => 
        if I_inputSampleValid = '1' then
            SR_futurState <= STORE1;
        else 
            SR_futurState <= WAIT_SAMPLE;
        end if;
        
      when STORE1 => 
            SR_futurState <= PROCESSING_LOOP1;
        
      when PROCESSING_LOOP1 =>
        if I_processingDone1 = '1' then
            SR_futurState <= STORE2;
        else 
            SR_futurState <= PROCESSING_LOOP1;
        end if;
       
      when STORE2 => 
            SR_futurState <= PROCESSING_LOOP2_1;
        
      when PROCESSING_LOOP2_1 =>
        if I_processingDone2_1 = '1' then
            SR_futurState <= PROCESSING_LOOP2_2;
        else 
            SR_futurState <= PROCESSING_LOOP2_1;
        end if;
        
      when PROCESSING_LOOP2_2 =>
        if I_processingDone2_2 = '1' then
            SR_futurState <= STORE3;
        else 
            SR_futurState <= PROCESSING_LOOP2_2;
        end if;
                
      when STORE3 => 
            SR_futurState <= PROCESSING_LOOP3;
        
      when PROCESSING_LOOP3 =>
        if I_processingDone3 = '1' then
            SR_futurState <= OUTPUT;
        else 
            SR_futurState <= PROCESSING_LOOP3;
        end if;
        
      when OUTPUT =>
        SR_futurState <= WAIT_END_SAMPLE;
        
      when others =>    ---WAIT_END_SAMPLE
        if I_inputSampleValid = '1' then
            SR_futurState <= WAIT_SAMPLE;
        else 
            SR_futurState <= WAIT_END_SAMPLE;
        end if;

    end case;
  end process;

  O_loadShift           <= "01" when SR_presentState = STORE1 else 
                           "10" when SR_presentState = STORE2 else 
                           "11" when SR_presentState = STORE3 else 
                           "00";
  O_initAddress         <= '1' when SR_presentState = STORE1 or SR_presentState = STORE2 or SR_presentState = STORE3 else '0';
  O_incrAddress         <= '1' when SR_presentState = PROCESSING_LOOP1 or SR_presentState = PROCESSING_LOOP2_1 or SR_presentState = PROCESSING_LOOP2_2 or SR_presentState = PROCESSING_LOOP3  else '0';
  O_initSum             <= '1' when SR_presentState = STORE1 or SR_presentState = STORE2 or SR_presentState = STORE3 else '0';
  O_loadSum             <= '1' when SR_presentState = PROCESSING_LOOP1 or SR_presentState = PROCESSING_LOOP2_1 or SR_presentState = PROCESSING_LOOP2_2 or SR_presentState = PROCESSING_LOOP3  else '0';
  O_loadOutput          <= '1' when SR_presentState = OUTPUT else '0' ;
  O_Selector            <= "01" when SR_presentState = STORE2 or  SR_presentState = PROCESSING_LOOP2_1 else 
                           "10" when SR_presentState = PROCESSING_LOOP2_2 else 
                           "11" when SR_presentState = STORE3 or SR_presentState = PROCESSING_LOOP3 else 
                           "00";
  --O_FilteredSampleValid <= '1' when SR_presentState = WAIT_END_SAMPLE else '0' ;

end architecture archi_operativeUnit;
